import copy
import dill

from compyl.__parser.finite_automaton import DFA
from compyl.__parser.rule_formatter import rules_are_valid, format_rules
from compyl.__parser.error import ParserError, ParserSyntaxError, ParserBuildError, GrammarError
from compyl.__parser.metaclass import MetaParser


__all__ = ['Parser', 'ParserError', 'ParserSyntaxError', 'ParserBuildError', 'GrammarError']


# ======================================================================================================================
# Parser Main Class
# ======================================================================================================================


class Parser(metaclass=MetaParser):
    def __init__(self, _dfa=None):
        # The following are class attributes generated by the metaclass
        self.rules = copy.deepcopy(self.__rules__)
        self.terminals = copy.deepcopy(self.__terminals__)

        if _dfa is not None:
            self.dfa = _dfa
        else:
            formatted_rules = format_rules(self.rules)
            self.dfa = DFA(formatted_rules, self.terminals)

    def __copy__(self):
        """
        Copy the parser reusing the same DFA
        :return:
        """
        dup = type(self)(_dfa=self.dfa)

        return dup

    def __deepcopy__(self, memo):
        """
        Copy the parser and its DFA
        :return:
        """

        dup = type(self)(_dfa=copy.deepcopy(self.dfa))

        return dup

    def reset(self):
        self.dfa.reset()

    def save(self, filename="lexer.p"):
        with open(filename, "wb") as file:
            dill.dump(self, file)

    @staticmethod
    def load(path):

        file = open(path, "rb")

        try:
            parser = dill.load(file)
        finally:
            file.close()

        if isinstance(parser, Parser):
            return parser
        else:
            raise ParserError("The unpickled object from " + path + " is not a Parser")

    def parse(self, token):
        if token:
            self.dfa.push(token)
            return None
        else:
            return self.end()

    def end(self):
        return self.dfa.end()

